;;; custom-load.el --- automatically extracted custom dependencies

;;; Code:

(autoload 'custom-add-loads "cus-load")

(custom-add-loads 'haskell '("haskell-doc" "haskell-font-lock" "haskell-ghci" "haskell-hugs" "haskell-indent" "inf-haskell"))
(custom-add-loads 'haskell-indent '("haskell-indent"))
(custom-add-loads 'haskell-ghci '("haskell-ghci"))
(custom-add-loads 'languages '("haskell-mode"))
(custom-add-loads 'haskell-hugs '("haskell-hugs"))

;;; custom-load.el ends here
